from .s3access import S3Access
from .s3path import S3Path
from .sql import EQ, LT, GT, NEQ, LTE, GTE, IN, AND, OR

__all__ = [
    'S3Access',
    'S3Path',

    'EQ',
    'LT',
    'GT',
    'NEQ',
    'LTE',
    'GTE',
    'IN',
    'AND',
    'OR',
]



from io import BytesIO
from numbers import Number
from typing import Union, Dict, Type, Sequence

import pandas as pd

from .dataframe import from_csv_bytes, merge_categories
from ..reader import Reader


class Pandas(Reader[pd.DataFrame]):
    def __init__(self, strict: bool = False):
        self._strict = strict

    def read(self, bs: Union[bytes, bytearray], columns: Dict[str, Union[Type, str]]) -> pd.DataFrame:
        if self._strict:
            return from_csv_bytes(bs, list(columns.keys()), columns)
        else:
            df = pd.read_csv(BytesIO(bs), header=None, names=columns.keys())
            for c, t in columns.items():
                if issubclass(t, Number):
                    df[c] = pd.to_numeric(df[c], errors='coerce')
            return df

    def combine(self, results: Sequence[pd.DataFrame]) -> pd.DataFrame:
        if not results:
            return pd.DataFrame([])
        if len(results) == 1:  # no need to concat, maybe return copy?
            return results[0]

        merge_categories(results)

        return pd.concat(results, ignore_index=True)

    @property
    def supports_caching(self):
        return True

    @property
    def serialization(self):
        return {'CSV': {
            'QuoteFields': 'ALWAYS',
            'QuoteEscapeCharacter': '"',
            'FieldDelimiter': ',',
            'QuoteCharacter': '"',
        }}

    def read_cache(self, cache_file: str) -> pd.DataFrame:
        return pd.read_parquet(cache_file)

    def write_cache(self, cache_file: str, contents: pd.DataFrame):
        contents.to_parquet(cache_file)



import asyncio
from typing import Sequence, Dict, Any, Optional, Tuple, Callable, AsyncIterator, Mapping, Coroutine
from aiobotocore.client import AioBaseClient

from ..select_spec import \
    Input, Output, DEFAULT_INPUT, DEFAULT_OUTPUT, DEFAULT_SCAN_RANGE, ScanRange, ExpressionType, \
    WITH_PROGRESS, NO_PROGRESS, simple_parquet_selection


ProgressCallback = Callable[[str, str, Any], None]
BucketKeyPair = Tuple[str, str]

"""
Sources are mappings where the mapping-key is the bucket and the value is a sequence of keys inside the bucket
Example: {'some-bucket': ['path/to/file.parquet', 'path/to/another_file.parquet', 'some/other/key.parquet']
"""
Sources = Mapping[str, Sequence[str]]


async def select(
        client: AioBaseClient,
        bucket: str,
        key: str,
        expression: str,
        expression_type: ExpressionType = ExpressionType.SQL,
        input_serialization: Input = DEFAULT_INPUT,
        output_serialization: Output = DEFAULT_OUTPUT,
        scan_range: ScanRange = DEFAULT_SCAN_RANGE,
        progress_callback: Optional[ProgressCallback] = None) -> bytes:
    obj = await client.select_object_content(
        Bucket=bucket,
        Key=key,
        Expression=expression,
        ExpressionType=expression_type.value,
        InputSerialization=input_serialization.params(),
        OutputSerialization=output_serialization.params(),
        ScanRange=scan_range.params(),
        RequestProgress=(WITH_PROGRESS if progress_callback else NO_PROGRESS).params()
    )
    res = bytearray()
    end = False
    async for event in obj['Payload']:
        if 'Records' in event:
            res.extend(event['Records']['Payload'])
        elif 'Progress' in event and progress_callback:
            progress_callback(bucket, key, event['Progress']['Details'])
        elif 'End' in event:
            end = True
    if not end:
        raise IOError("stream was not read in full")
    return res


def _select_tasks(
        client: AioBaseClient,
        sources: Sources,
        expression: str,
        expression_type: ExpressionType = ExpressionType.SQL,
        input_serialization: Input = DEFAULT_INPUT,
        output_serialization: Output = DEFAULT_OUTPUT,
        scan_range: ScanRange = DEFAULT_SCAN_RANGE,
        progress_callback: Optional[ProgressCallback] = None) -> Tuple[Sequence[Coroutine], Sequence[BucketKeyPair]]:
    tasks = []
    key_list = []
    for bucket, keys in sources.items():
        for key in keys:
            tasks.append(
                select(client, bucket, key, expression, expression_type, input_serialization,
                       output_serialization, scan_range, progress_callback))
            key_list.append((bucket, key))
    return tasks, key_list


async def multiple(
        client: AioBaseClient,
        sources: Sources,
        expression: str,
        expression_type: ExpressionType = ExpressionType.SQL,
        input_serialization: Input = DEFAULT_INPUT,
        output_serialization: Output = DEFAULT_OUTPUT,
        scan_range: ScanRange = DEFAULT_SCAN_RANGE,
        progress_callback: Optional[ProgressCallback] = None) -> Dict[BucketKeyPair, bytes]:
    tasks, key_list = _select_tasks(
        client, sources, expression, expression_type, input_serialization,
        output_serialization, scan_range, progress_callback)
    results = await asyncio.gather(*tasks)
    return dict(zip(key_list, results))


async def multiple_as_completed(
        client: AioBaseClient,
        sources: Sources,
        expression: str,
        expression_type: ExpressionType = ExpressionType.SQL,
        input_serialization: Input = DEFAULT_INPUT,
        output_serialization: Output = DEFAULT_OUTPUT,
        scan_range: ScanRange = DEFAULT_SCAN_RANGE,
        progress_callback: Optional[ProgressCallback] = None) -> AsyncIterator[Tuple[bytes, BucketKeyPair]]:
    tasks, key_list = _select_tasks(
        client, sources, expression, expression_type, input_serialization,
        output_serialization, scan_range, progress_callback)
    for i, file_f in enumerate(asyncio.as_completed(tasks)):
        file = await file_f
        yield file, key_list[i]


async def parquets_to_csv_bytes(
        client: AioBaseClient,
        sources: Sources,
        columns: Optional[Sequence[str]] = None,
        filters: Sequence[Tuple[Any, str, Any]] = None,
        progress_callback: Optional[ProgressCallback] = None) -> bytes:
    """
    example usage:
    -----------------------------------------
    def print_progress(bucket, key, details):
        if details == 'completed':
            print(bucket, key, details)
    bucket = 'some-bucket'
    keys = ['key/key1.parquet', 'key/key2.parquet', ...]
    columns = ['col1', 'col2', 'col5', 'col10']  # some columns all files share
    filters = [('col1', '=', 'some string value'), ('col5', '>=', 10.2), ('col5', '<', 20.3)]
    session = aiobotocore.get_session()
    async with session.create_client('s3') as client:
        blob = await parquets_to_csv_bytes(client, {bucket: keys}, columns, filters, print_progress)
    content = io.StringIO(blob.decode('utf-8'))
    for row in csv.reader(content):
        do_something_with_row(row)
    :param client: initialized AioBaseClient to use
    :param sources: a dictionary mapping a bucket to a list of files to scan from it. ALL files must be in the same
                    schema
    :param columns: which column to load
    :param filters: how to filter the parquet file
    :param progress_callback: a callback to report progress to, each file when completed will also report completion
    :return: bytes
    """
    q = simple_parquet_selection(columns, filters)
    total = bytearray()
    async for content, (bucket, key) in multiple_as_completed(
            client, sources, q, progress_callback=progress_callback):
        total.extend(content)
        if progress_callback:
            progress_callback(bucket, key, "completed")
    return total



from s3access.s3async.s3select import select, multiple, multiple_as_completed, parquets_to_csv_bytes

__all__ = [
    'select',
    'multiple',
    'multiple_as_completed',
    'parquets_to_csv_bytes'
]



import base64
import fnmatch
import glob
import hashlib
import logging
import multiprocessing
import os
import sys
import threading
from concurrent import futures
from dataclasses import dataclass, field
from typing import Union, List, Iterator, Dict, Type, TypeVar, Optional, Tuple

import boto3
from botocore.client import BaseClient
from readstr import readstr

from .reader import Reader
from .s3path import S3Path
from .sql import make_conditions, Filter, FilterResolved, SimpleFilter, AND, OR

logger = logging.getLogger(__name__)

try:
    # noinspection PyUnresolvedReferences
    from s3access.s3pandas.reader import Pandas

    DEFAULT_READER = Pandas()
except ImportError:
    logger.warning("could not load pandas reader, using Python default reader")
    from s3access.reader import Python

    DEFAULT_READER = Python()


@dataclass
class LsResult:
    prefixes: List[S3Path] = field(default_factory=list)
    contents: List[S3Path] = field(default_factory=list)

    @property
    def directories(self):
        return self.prefixes

    @property
    def files(self):
        return self.contents


# noinspection PyShadowingBuiltins
def read_value(value, type):
    if value is None:
        return None
    try:
        return readstr(value, type)
    except ValueError:
        return None


T = TypeVar('T')


class _NoValue:
    pass


def build_simple(fs: List[SimpleFilter], combiner) -> str:
    cs = []
    for f in fs:
        if isinstance(f, AND):
            c = build_simple(f.conditions, 'AND')
            if c:
                cs.append(f'({c})')
        elif isinstance(f, OR):
            c = build_simple(f.conditions, 'OR')
            if c:
                cs.append(f'({c})')
        else:
            r, o, v = f
            if o in ('=', '=='):
                o = '='
            if o in ('!=', '<>', '/='):
                o = '!='
            if isinstance(v, str):
                v.replace("'", "''")
                v = f"'{v}'"
            cs.append(f'("{r}" {o} {v})')
    return f" {combiner} ".join(cs)


def build_expression(s3path: S3Path, columns: Dict[str, Type], filters: FilterResolved) -> str:
    # noinspection SqlResolve,SqlNoDataSourceInspection
    query = f"SELECT {', '.join(f's.{key}' for key in columns.keys())} FROM S3Object s"

    if isinstance(filters, dict):
        object_filters = {k: f for k, f in filters.items() if k not in s3path.params}
        if object_filters:
            query += ' WHERE '
            query += ' AND '.join(c.get_sql_fragment(f"s.{k}") for k, c in object_filters.items())
    else:
        fs = build_simple(filters, 'AND')
        if fs:
            query += f" WHERE {fs}"
    return query


class S3Access:
    def __init__(self, parallelism: int = multiprocessing.cpu_count() * 4, cachedir: Optional[str] = _NoValue):
        self._num_workers: int = parallelism
        if cachedir is _NoValue:
            cachedir = os.getenv('S3ACCESSCACHE')
        self._cachedir: Optional[str] = cachedir

    @staticmethod
    def s3client() -> BaseClient:
        """
        Returns an S3 client for this thread.
        """
        try:
            client_ = threading.local().s3client
        except AttributeError:
            endpoint_url = os.getenv('S3_ENDPOINT_URL') or None  # prevent empty string
            client_ = boto3.client('s3', endpoint_url=endpoint_url)
            threading.local().s3client = client_
        return client_

    def ls(self, s3path: Union[str, S3Path]) -> LsResult:
        """
        Lists all files and folders at the given location.

        The s3path given is assumed to refer to a directory in S3.
        Yes, there are no directories in S3 but common prefixes, but
        that does not stop anyone from thinking about it like directories.

        The result is separated into the prefixes (directories) and
        contents (files).

        This method is faster than just listing everything with the same
        prefix, due to the use of `Delimiter` in the api call. If there are
        more than 1000 entries this method automatically follows up using
        the continuation token (this is hidden away from the user).
        """
        if isinstance(s3path, str):
            s3path = S3Path(s3path)
        key = s3path.key
        if key and key[-1] != '/':
            key += '/'
        response = self.s3client().list_objects_v2(
            Bucket=s3path.bucket,
            Prefix=key,
            Delimiter='/',
        )
        result = LsResult()
        result.prefixes.extend(s3path.with_key(common_prefix['Prefix'])
                               for common_prefix in response.get('CommonPrefixes', []))
        result.contents.extend(s3path.with_key(content['Key'])
                               for content in response.get('Contents', []))
        while 'NextContinuationToken' in response:
            response = self.s3client().list_objects_v2(
                Bucket=s3path.bucket,
                Prefix=s3path.key + '/',
                Delimiter='/',
                ContinuationToken=response['NextContinuationToken']
            )
            result.prefixes.extend(s3path.with_key(common_prefix['Prefix'])
                                   for common_prefix in response.get('CommonPrefixes', []))
            result.contents.extend(s3path.with_key(content['Key'])
                                   for content in response.get('Contents', []))
        return result

    def ls_prefix(self, s3path: Union[str, S3Path]) -> Iterator[S3Path]:
        """
        List all entries with the given s3path as prefix.

        Do so exhaustively, i.e. if the number of entries exceeds the maximum page size
        this method will automatically follow up using a contiuation token. This happens
        on-demand as it returns an iterator, i.e. the seconds api call will only happen
        once you consumed the first 1000 entries from the iterator returned.
        """
        if isinstance(s3path, str):
            s3path = S3Path(s3path)
        response = self.s3client().list_objects_v2(
            Bucket=s3path.bucket,
            Prefix=s3path.key,
        )
        yield from (s3path.with_key(content['Key']) for content in response.get('Contents', []))
        while 'NextContinuationToken' in response:
            logger.debug("Looking for more items using %s", response['NextContinuationToken'])
            response = self.s3client().list_objects_v2(
                Bucket=s3path.bucket,
                Prefix=s3path.key,
                ContinuationToken=response['NextContinuationToken']
            )
            yield from (s3path.with_key(content['Key']) for content in response.get('Contents', []))

    @staticmethod
    def is_glob(value):
        return glob.escape(value) != value

    def glob(self, s3path: Union[str, S3Path]) -> Iterator[S3Path]:
        """
        Lists all entries that match a given glob-pattern.

        glob-patterns are actually more powerful than just the asterisk-any-match,
        but this method is concerned with those only.

        TODO: if needed pimp this in the future. Currently this method splits the
        path on the first occurence of an asterisk, fetches everything with the
        determined prefix, and checks the results for whether they match. No optimization
        is applied if other glob patterns (?, []) are used.

        Behind the scenes this method uses ls_prefix which lazily fetches more items
        using a continuation token if needed, and this laziness propagates into this
        method, which also does return an Iterator.
        """
        if isinstance(s3path, str):
            s3path = S3Path(s3path)

        if '*' in s3path.key:
            prefix, _r = s3path.key.split('*', maxsplit=1)
        else:
            prefix = s3path.key
        for p in self.ls_prefix(s3path.with_key(prefix)):
            if fnmatch.fnmatch(p.key, s3path.key):
                yield p

    @staticmethod
    def _check_path(path: S3Path, filters: FilterResolved) -> bool:
        if isinstance(filters, dict):
            for k, c in filters.items():
                if k in path.params:
                    if not c.check(path[k]):
                        return False
        else:
            def chk(sf: SimpleFilter) -> bool:
                if isinstance(sf, AND):
                    return all(chk(cn) for cn in sf.conditions)
                elif isinstance(sf, OR):
                    return any(chk(cn) for cn in sf.conditions)
                else:
                    r, o, v = sf
                    if r not in path.params:
                        return True
                    r = path.params[r]
                    if isinstance(v, int):
                        r = int(r)
                    elif isinstance(v, float):
                        r = float(r)
                    if o in ('=', '=='):
                        return r == v
                    elif o in ('!=', '<>', '/='):
                        return r != v
                    elif o == '>':
                        return r > v
                    elif o == '<':
                        return r < v
                    elif o == '>=':
                        return r >= v
                    elif o == '<=':
                        return r <= v

            for f in filters:
                if not chk(f):
                    return False
        return True

    def _select_glob(self,
                     s3path: Union[str, S3Path],
                     columns: Dict[str, Type],
                     filters: FilterResolved,
                     reader: Reader[T]) -> T:
        paths = self.glob(s3path)

        pool = futures.ThreadPoolExecutor(max_workers=self._num_workers)

        def worker(p: S3Path) -> T:
            logger.debug("Spawned selecting from %s", p)
            result = self.select(p, columns, filters, reader)
            return result

        it = iter(paths)
        pending_futures = []
        results: List[T] = []
        try:
            spawned = 0
            while True:  # the loop will stop when next() raises StopIteration
                # poor man's barrier
                while spawned < self._num_workers:
                    path = next(it)
                    if filters and not self._check_path(path, filters):
                        continue
                    spawned += 1
                    pending_futures.append(pool.submit(worker, path))
                done, pending = futures.wait(pending_futures, return_when=futures.FIRST_COMPLETED)
                spawned -= len(done)
                results.extend(future.result() for future in done)
                pending_futures = [*pending]
        except StopIteration:
            done, _ = futures.wait(pending_futures, return_when=futures.ALL_COMPLETED)
            results.extend(future.result() for future in done)
        finally:
            pool.shutdown()
        return reader.combine(results)

    def _select(self,
                s3path: Union[str, S3Path],
                columns: Dict[str, Type],
                query: str,
                reader: Reader[T]) -> T:

        logger.debug("Issuing S3 Select Query: ``%s'' on %s", query, s3path)
        response = self.s3client().select_object_content(
            Bucket=s3path.bucket,
            Key=s3path.key,
            InputSerialization={'Parquet': {}},
            OutputSerialization=reader.serialization,
            ExpressionType='SQL',
            Expression=query,
        )
        bs = self._read_s3_select_response(response)
        return reader.read(bs, columns=columns)

    @staticmethod
    def _read_s3_select_response(response) -> bytearray:
        res = bytearray()
        for stream in response['Payload']:
            try:
                res.extend(stream['Records']['Payload'])
            except KeyError:
                continue
        return res

    @staticmethod
    def _cache_fingerprint(s3path: S3Path, query: str) -> str:
        md = hashlib.sha1()
        md.update(str(s3path).encode('utf8'))
        md.update(query.encode('utf8'))
        fingerprint: str = base64.b32encode(md.digest()).decode('ascii')
        return fingerprint

    def _realm(self, s3path: S3Path) -> str:
        if not self.is_glob(s3path.key):
            return 'file'
        return 'glob_' + s3path.key.translate(str.maketrans({'/': ',', '*': '_', '[': '_', ']': '_', '?': '_'}))

    def _in_cache(
              self,
              s3path: Union[str, S3Path],
              query: str) -> Tuple[bool, Optional[str]]:
        fingerprint = self._cache_fingerprint(s3path, query)
        realm = self._realm(s3path)
        cache_file = None
        if self._cachedir and os.path.isdir(self._cachedir):
            cache_file = f"{self._cachedir}/{s3path.bucket}/{realm}/{fingerprint}.bin"
            if os.path.exists(cache_file):
                return True, cache_file
        return False, cache_file

    def select(self,
               s3path: Union[str, S3Path],
               columns: Dict[str, Type],
               filters: Optional[Filter] = None,
               reader: Reader[T] = DEFAULT_READER
               ) -> T:
        """
        Selects the given columns of the given type from the given s3path.

        Args:
            s3path: A url to s3 like s3://bucket/path/file.parquet or s3://bucket/date=2020-*/*.parquet
            columns: A dict of name/type pairs for which columns to select. For example:
                {
                    'campaign_name': str,
                    'campaign_id': str,
                    'cost': float,
                    'installs': int,
                }
            filters: An optional dict of filters which are combined using AND. Each key refers to a column,
                each value is a Condition of which there are EQ, LT, LTE, GT, GTE, NET, and IN. For example:
                {
                    'cost': GT(2.5),
                    'country': IN('USA', 'NZL', 'ITA'),
                }
            reader: A Reader that deserializes the response. Defaults to the Pandas dataframe reader.
        """
        if isinstance(s3path, str):
            s3path = S3Path(s3path)
        if isinstance(filters, dict):
            filters = make_conditions(filters)
        elif not isinstance(filters, list):
            filters = {}

        query = build_expression(s3path, columns, filters)

        in_cache, cache_file = False, None
        # noinspection PyBroadException
        try:
            if reader.supports_caching:
                in_cache, cache_file = self._in_cache(s3path, query)
                if in_cache:
                    return reader.read_cache(cache_file)
        except Exception as exc:
            logger.warning("Could not load cache file %s due to %s (removing file)",
                           cache_file, type(exc).__name__, exc_info=sys.exc_info())
            # noinspection PyBroadException
            try:
                os.remove(cache_file)
            except Exception as exc:
                logger.warning("Removing broken cache file %s failed due to %s",
                               cache_file, type(exc).__name__, exc_info=sys.exc_info())

        if self.is_glob(s3path.key):
            result = self._select_glob(s3path, columns, filters, reader)
        else:
            result = self._select(s3path, columns, query, reader)

        if reader.supports_caching and cache_file:
            cachedir = os.path.join(cache_file.rpartition('/')[0])
            try:
                os.makedirs(cachedir, exist_ok=True)
            except OSError:
                logger.warning("could not create cache directory %s, not caching", cachedir)
            else:
                logger.debug("Writing result to cache %s", cache_file)
                reader.write_cache(cache_file, result)

        return result

    async def select_async(
              self,
              s3path: Union[str, S3Path],
              columns: Dict[str, Type],
              filters: Optional[Filter] = None,
              reader: Reader[T] = DEFAULT_READER) -> T:
        # async interface is optional
        import aiobotocore
        from s3access.s3async.s3select import multiple_as_completed, Output

        if isinstance(s3path, str):
            s3path = S3Path(s3path)
        if isinstance(filters, dict):
            filters = make_conditions(filters)
        elif not isinstance(filters, list):
            filters = {}

        query = build_expression(s3path, columns, filters)

        is_glob = self.is_glob(s3path.key)
        in_cache, global_cache_file = False, None

        if is_glob and reader.supports_caching:
            in_cache, global_cache_file = self._in_cache(s3path, query)
            if in_cache:
                return reader.read_cache(global_cache_file)

        # TODO: the glob part should eventually be asynchronous as well
        paths = [s3path] if not is_glob else self.glob(s3path)
        if not paths:
            return reader.combine([])

        results = []
        missing_paths = []
        cache_files: Dict[Tuple[str, str], str] = {}
        for p in paths:
            in_cache, cache_file = self._in_cache(p, query)
            if in_cache:
                results.append(reader.read_cache(cache_file))
            else:
                missing_paths.append(p)
                if cache_file:
                    cache_files[(p.bucket, p.key)] = cache_file
        if len(missing_paths) == 0:
            combined = reader.combine(results)
            if is_glob and global_cache_file:
                reader.write_cache(global_cache_file, combined)
            return combined

        bucket = missing_paths[0].bucket
        sources = {bucket: [p.key for p in missing_paths]}
        query = build_expression(s3path, columns, filters)
        session = aiobotocore.get_session()
        results = []
        async with session.create_client('s3') as client:
            async for content, cache_key in multiple_as_completed(
                      client, sources, query, output_serialization=Output(reader.serialization)):
                logger.debug("fetch completed for %s - %s", cache_key[0], cache_key[1])
                parsed = reader.read(content, columns)
                results.append(parsed)
                cache_file = cache_files.get(cache_key)
                if cache_file:
                    reader.write_cache(cache_file, content)

        combined = reader.combine(results)
        if is_glob and global_cache_file:
            reader.write_cache(global_cache_file, combined)

        return combined



import io
from datetime import date, datetime
from typing import Sequence, Optional, Dict, Any

import pandas as pd
from pandas.api.types import union_categoricals


def from_csv_bytes(
        data: bytes,
        columns: Sequence[str],
        types: Optional[Dict[str, Any]] = None,
        encoding: str = 'utf-8') -> Optional[pd.DataFrame]:
    if len(data) == 0:
        return None
    parse_dates = None
    if types:
        parse_dates = [k for k, v in types.items() if v is date or v is datetime]
        types = {k: v for k, v in types.items() if v is not date and v is not datetime}
    file_handle = io.BytesIO(data)
    res = pd.read_csv(
        file_handle,
        header=None,
        encoding=encoding,
        names=columns,
        dtype=types,
        parse_dates=parse_dates,
        infer_datetime_format=True)
    return res


def merge_categories(results: Sequence[pd.DataFrame]):
    to_cat = []
    df = results[0]
    for col in df:
        if isinstance(df[col].dtype, pd.CategoricalDtype):
            to_cat.append(col)
    if to_cat:
        unify = {}
        for col in to_cat:
            u = unify[col] = []
            for result in results:
                u.append(result[col])
        unified = {col: union_categoricals(v).categories for col, v in unify.items()}

        for res in results:
            for col in to_cat:
                res[col] = pd.Categorical(res[col].values, categories=unified[col])



from __future__ import annotations

from abc import abstractmethod, ABC
from collections.abc import Collection
from numbers import Number
from typing import Union, Sequence, Dict, List, Tuple, Literal



def quote(value) -> str:
    if value is None:
        return "NULL"
    if isinstance(value, Collection) and not isinstance(value, str):
        return f"({', '.join(quote(v) for v in value)})"
    if isinstance(value, Number):
        return f"{value}"
    value = str(value).replace("'", "''")
    return f"'{value}'"


class Condition(ABC):
    def __init__(self, value):
        self.__value = value

    @abstractmethod
    def get_sql_fragment(self, ref) -> str:
        raise NotImplementedError

    @abstractmethod
    def check(self, ref) -> bool:
        raise NotImplementedError

    @property
    def _quoted(self) -> str:
        return quote(self._value)

    @property
    def _value(self):
        return self.__value

    def __str__(self) -> str:
        return f"{type(self).__name__}({self._quoted})"


class EQ(Condition):
    def check(self, ref) -> bool:
        if self._value is None:
            return ref is None
        return ref == self._value

    def get_sql_fragment(self, ref):
        if self._value is None:
            return f"{ref} IS NULL"
        return f"{ref} = {self._quoted}"


class LT(Condition):
    def check(self, ref) -> bool:
        return ref < self._value

    def get_sql_fragment(self, ref):
        return f"{ref} < {self._quoted}"


class GT(Condition):
    def check(self, ref) -> bool:
        return ref > self._value

    def get_sql_fragment(self, ref):
        return f"{ref} > {self._quoted}"


class NEQ(Condition):
    def check(self, ref) -> bool:
        if self._value is None:
            return ref is not None
        return ref != self._value

    def get_sql_fragment(self, ref):
        if self._value is None:
            return f"{ref} IS NOT NULL"
        return f"{ref} <> {self._quoted}"


class LTE(Condition):
    def check(self, ref) -> bool:
        return ref <= self._value

    def get_sql_fragment(self, ref):
        return f"{ref} <= {self._quoted}"


class GTE(Condition):
    def check(self, ref) -> bool:
        return ref >= self._value

    def get_sql_fragment(self, ref):
        return f"{ref} >= {self._quoted}"


class IN(Condition):
    def __init__(self, *values):
        super().__init__(values)

    def check(self, ref) -> bool:
        return ref in self._value

    def get_sql_fragment(self, ref):
        return f"{ref} IN {self._quoted}"


class AND(Condition):
    def __init__(self, *conditions: Union[Condition, SimpleFilter]):
        super().__init__(None)
        self._conditions = conditions

    @property
    def conditions(self):
        return self._conditions

    def check(self, ref) -> bool:
        for condition in self._conditions:
            if not condition.check(ref):
                return False
        return True

    def get_sql_fragment(self, ref) -> str:
        return ' AND '.join(f"({c.get_sql_fragment(ref)})" for c in self._conditions)


class OR(Condition):
    def __init__(self, *conditions: Union[Condition, SimpleFilter]):
        super().__init__(None)
        self._conditions = conditions

    @property
    def conditions(self):
        return self._conditions

    def check(self, ref) -> bool:
        for condition in self._conditions:
            if condition.check(ref):
                return True
        return False

    def get_sql_fragment(self, ref) -> str:
        return ' OR '.join(f"({c.get_sql_fragment(ref)})" for c in self._conditions)


Conditionable = Union[Condition, str, int, float, Sequence[str]]

Operator = Literal['>=', '>', '<=', '<', '=', '==', '<>', '!=', '/=']

Value = Union[str, int, float]

SimpleFilter = Union[AND, OR, Tuple[str, str, Value]]

Filter = Union[Dict[str, Conditionable], List[SimpleFilter]]

FilterResolved = Union[Dict[str, Condition], List[SimpleFilter]]


def make_condition(v):
    if isinstance(v, (str, int, float)):
        return EQ(v)
    if isinstance(v, Sequence):
        return IN(*v)
    assert isinstance(v, Condition), "must be a condition"
    return v


def make_conditions(conditions: Dict[str, Conditionable]) -> Dict[str, Condition]:
    return {k: make_condition(v) for k, v in conditions.items()}



from typing import Mapping, Sequence, Optional, Dict, Any, Tuple

import pandas as pd
from aiobotocore.client import AioBaseClient

from .s3select import ProgressCallback, parquets_to_csv_bytes
from ..s3pandas.dataframe import from_csv_bytes


async def dataframe_from_parquets(
        client: AioBaseClient,
        sources: Mapping[str, Sequence[str]],
        columns: Optional[Sequence[str]] = None,
        types: Optional[Dict[str, Any]] = None,
        filters: Sequence[Tuple[Any, str, Any]] = None,
        progress_callback: Optional[ProgressCallback] = None) -> Optional[pd.DataFrame]:
    """
    example usage:
    -----------------------------------------
    def print_progress(bucket, key, details):
        if details == 'completed':
            print(bucket, key, details)

    bucket = '<some bucket you have access to>'
    keys = [<parquet file key path>, ...]
    columns = ['col1', 'col2', 'col5', 'col10']  # some columns all files share
    dtypes = {'col1': str, 'col2': 'category', 'col5': float, 'col10': date}
    filters = [('col1', '=', 'some string value'), ('col5', '>=', 10.2), ('col5', '<', 20.3)]
    session = aiobotocore.get_session()
    async with session.create_client('s3') as client:
        new_df = await select_dataframe_from_parquets(client, {bucket: keys}, columns, dtypes, filters, print_progress)
    :param client: initialized AioBaseClient to use
    :param sources: a dictionary mapping a bucket to a list of files to scan from it. ALL files must be in the same
                    schema
    :param columns: which column to load
    :param types: type conversion to the resulting columns
    :param filters: how to filter the parquet file
    :param progress_callback: a callback to report progress to, each file when completed will also report completion
    :return: pandas.DataFrame
    """
    blob = await parquets_to_csv_bytes(client, sources, columns, filters, progress_callback)
    return from_csv_bytes(blob, columns, types)



from __future__ import annotations

import re
from typing import Optional, Dict


class S3Path:
    _path_pattern = re.compile(r"([sS]3[aAnN]?:/*)*(?P<bucket>[^/:]+)(/(?P<key>.+)?)?")

    @property
    def bucket(self) -> str:
        return self._bucket

    @property
    def key(self) -> str:
        return self._key

    @property
    def params(self) -> Dict[str, str]:
        return self._params

    def with_bucket(self, new_bucket) -> S3Path:
        return S3Path(new_bucket, self._key)

    def with_key(self, new_key) -> S3Path:
        return S3Path(self._bucket, new_key)

    def get_url(self, protocol: str = 's3'):
        return f"{protocol}://{self.bucket}/{self.key}"

    def with_params(self, **kwargs) -> S3Path:
        components = []
        applied = set()
        for component in self._key.split('/'):
            if not component:
                continue
            parts = component.split('=', 1)
            if len(parts) == 2:
                key, value = parts
                if key in kwargs:
                    applied.add(key)
                    components.append(f"{key}={kwargs[key]}")
                    continue
            components.append(component)
        for key, value in kwargs.items():
            if key not in applied:
                components.append(f"{key}={value}")
        return self.with_key('/'.join(components))

    def __init__(self, bucket_or_url, key: Optional[str] = None):
        if isinstance(bucket_or_url, S3Path):
            self._bucket = bucket_or_url._bucket
            self._key = bucket_or_url._key
        elif bucket_or_url and key is not None:
            self._bucket = bucket_or_url
            self._key = key
        else:
            result = self._path_pattern.fullmatch(bucket_or_url)
            if not result:
                raise ValueError
            self._bucket = result.group('bucket') or ''
            self._key = result.group('key') or ''
        self._key = self._key.rstrip('/')
        self._params = {}
        for component in self._key.split('/'):
            parts = component.split('=', 1)
            if len(parts) == 2:
                key, value = parts
                self._params[key] = value

    def __getitem__(self, item):
        return self.params[item]

    def __contains__(self, item):
        return item in self.params

    def __truediv__(self, key) -> S3Path:
        return self.with_key(self.key + '/' + key)

    def __eq__(self, other):
        if isinstance(other, S3Path):
            return self._bucket == other._bucket and self._key == other._key
        if isinstance(other, str):
            return self == S3Path(other)
        raise NotImplementedError

    def __hash__(self):
        return hash((self._bucket, self._key))

    def __str__(self):
        return f's3://{self._bucket}/{self._key}'

    def __repr__(self):
        return f'S3Path({str(self)})'



import dataclasses
from datetime import date, datetime
from enum import Enum
from typing import Dict, Optional, Union, Sequence, Tuple, Any


# https://docs.aws.amazon.com/AmazonS3/latest/API/API_SelectObjectContent.html


class ExpressionType(Enum):
    SQL = 'SQL'


class InputSerialization(Enum):
    PARQUET = 'Parquet'
    CSV = 'CSV'
    JSON = 'JSON'


class OutputSerialization(Enum):
    CSV = 'CSV'
    JSON = 'JSON'


class QuoteField(Enum):
    ASNEEDED = 'ASNEEDED'
    ALWAYS = 'ALWAYS'


@dataclasses.dataclass(frozen=True)
class RequestProgress:
    enabled: bool = False

    def params(self):
        return {'Enabled': self.enabled}


NO_PROGRESS = RequestProgress()
WITH_PROGRESS = RequestProgress(True)


@dataclasses.dataclass(frozen=True)
class ParquetInput:

    @staticmethod
    def params():
        return {InputSerialization.PARQUET.value: {}}


DEFAULT_PARQUET_INPUT = ParquetInput()


class FileHeaderInfo(Enum):
    NONE = 'NONE'
    IGNORE = 'IGNORE'
    USE = 'USE'


@dataclasses.dataclass(frozen=True)
class CsvInput:
    """
    setting 'allow_quoted_record_delimiter' to True may lower performance.
    """
    allow_quoted_record_delimiter: bool = False
    comments: Optional[str] = None
    field_delimiter: str = ','
    file_header_info: FileHeaderInfo = FileHeaderInfo.NONE
    quote_character: str = '"'
    quote_escape_character: str = '"'
    record_delimiter: Optional[str] = None

    def params(self):
        kwargs = {
            'AllowQuotedRecordDelimiter': self.allow_quoted_record_delimiter,
            'FieldDelimiter': self.field_delimiter,
            'FileHeaderInfo': self.file_header_info.value,
            'QuoteCharacter': self.quote_character,
            'QuoteEscapeCharacter': self.quote_escape_character
        }
        if self.comments is not None:
            kwargs['Comments'] = self.comments
        if self.record_delimiter is not None:
            kwargs['RecordDelimiter'] = self.record_delimiter
        return {InputSerialization.CSV.value: kwargs}


DEFAULT_CSV_INPUT = CsvInput()


class JsonType(Enum):
    DOCUMENT = 'DOCUMENT'
    LINES = 'LINES'


@dataclasses.dataclass(frozen=True)
class JsonInput:
    type: Optional[JsonType] = None

    @staticmethod
    def serialization_kind() -> InputSerialization:
        return InputSerialization.JSON

    def params(self):
        kwargs = {}
        if self.type is not None:
            kwargs['Type'] = self.type.value
        return {InputSerialization.JSON.value: kwargs}


DEFAULT_JSON_INPUT = JsonInput()

AnyInput = Union[ParquetInput, CsvInput, JsonInput, Dict[str, Any]]


class CompressionType(Enum):
    NONE = 'NONE'
    GZIP = 'GZIP'
    BZIP = 'BZIP'


@dataclasses.dataclass(frozen=True)
class Input:
    compression_type: CompressionType = CompressionType.NONE
    input: AnyInput = DEFAULT_PARQUET_INPUT

    def params(self):
        res = self.input if isinstance(self.input, dict) else self.input.params()
        res['CompressionType'] = self.compression_type.value
        return res


DEFAULT_INPUT = Input()


@dataclasses.dataclass(frozen=True)
class CsvOutput:
    quote_fields: QuoteField = QuoteField.ALWAYS
    record_delimited: str = '\n'
    field_delimiter: str = ','
    quote_character: str = '"'
    quote_escape_character = '"'

    def params(self):
        return {
            OutputSerialization.CSV.value: {
                'QuoteFields': self.quote_fields.value,
                'RecordDelimiter': self.record_delimited,
                'FieldDelimiter': self.field_delimiter,
                'QuoteCharacter': self.quote_character,
                'QuoteEscapeCharacter': self.quote_escape_character
            }
        }


CSV_DEFAULT_OUTPUT = CsvOutput()


@dataclasses.dataclass(frozen=True)
class JsonOutput:
    record_delimiter: str = '\n'

    def params(self):
        return {OutputSerialization.JSON.value: {'RecordDelimiter': self.record_delimiter}}


JSON_DEFAULT_OUTPUT = JsonOutput()

AnyOutput = Union[CsvOutput, JsonOutput, Dict[str, Any]]


@dataclasses.dataclass(frozen=True)
class Output:
    output: AnyOutput = CSV_DEFAULT_OUTPUT

    def params(self):
        return self.output if isinstance(self.output, dict) else self.output.params()


DEFAULT_OUTPUT = Output()


@dataclasses.dataclass(frozen=True)
class ScanRange:
    start: int = 0
    end: Optional[int] = None

    def params(self) -> Dict[str, int]:
        res = {'Start': self.start}
        if self.end is not None:
            res['End'] = self.end
        return res


DEFAULT_SCAN_RANGE = ScanRange()


def _quote(v) -> str:
    if v is None:
        return 'NULL'
    if isinstance(v, str):
        v = v.replace("'", "''")
        return f"'{v}'"
    if isinstance(v, (date, datetime)):
        return f"'{v.isoformat()}'"
    if isinstance(v, Sequence):
        return f"({','.join(map(_quote, v))})"
    return str(v)


def simple_parquet_selection(
        columns: Optional[Sequence[str]] = None,
        filters: Sequence[Tuple[Any, str, Any]] = None) -> str:
    """
    select the given columns and filter the data by using "AND" between all given filters

    example usage:
    ---------------------------------------------
    columns = ['col1', 'col2', 'col5', 'col10']  # some columns all files share
    filters = [('col1', '=', 'some string value'), ('col5', '>=', 10.2), ('col5', '<', 20.3)]

    output:
    SELECT s.col1, s.col2, s.col5, s.col10 FROM S3Object s WHERE s.col1 = 'some string value' AND s.col5 >= 10.2 AND s.col5 < 20.3

    :param columns: the columns to select from the parquet
    :param filters: how to filter the data
    :return: s3-select query string
    """
    projection = '*' if not columns else ','.join(f"s.{col}" for col in columns)
    q = f'SELECT {projection} FROM S3Object s'
    if filters:
        f = ' AND '.join(f"s.{col} {op} {_quote(right)}" for col, op, right in filters)
        q = f'{q} WHERE {f}'
    return q



import csv
from abc import abstractmethod, ABC
from io import StringIO
from json import JSONDecoder, JSONDecodeError
from typing import Generic, TypeVar, Union, Dict, Type, Iterator, Sequence, List

R = TypeVar('R')


class Reader(ABC, Generic[R]):
    @abstractmethod
    def read(self, bs: Union[bytes, bytearray], columns: Dict[str, Type]) -> R:
        raise NotImplementedError

    @abstractmethod
    def combine(self, results: Sequence[R]) -> R:
        raise NotImplementedError

    @property
    @abstractmethod
    def serialization(self):
        raise NotImplementedError

    @property
    def supports_caching(self):
        return False

    def read_cache(self, cache_file: str) -> R:
        return NotImplemented

    def write_cache(self, cache_file: str, contents: R):
        return NotImplemented


RowDict = Sequence[Union[int, float, str]]


class Python(Reader[Iterator[RowDict]]):
    def read(self, bs: Union[bytes, bytearray], columns: Dict[str, Type]) -> Iterator[RowDict]:
        yield from csv.reader(StringIO(bs.decode('utf8')), dialect='unix')

    def combine(self, results: Sequence[Iterator[RowDict]]) -> Iterator[RowDict]:
        for result in results:
            yield from result

    @property
    def serialization(self):
        return {'CSV': {
            'QuoteFields': 'ALWAYS',
            'QuoteEscapeCharacter': '"',
            'FieldDelimiter': ',',
            'QuoteCharacter': '"',
        }}


JsonValue = Union[None, bool, str, int, float, List['JSON'], Dict[str, 'JSON']]


class Json(Reader[Iterator[JsonValue]]):
    def read(self, bs: Union[bytes, bytearray], columns: Dict[str, Type]) -> Iterator[JsonValue]:
        s = bs.decode('utf8')
        decoder = JSONDecoder()
        offset = 0
        try:
            while True:
                obj, offset = decoder.raw_decode(s, offset)
                offset += 1  # advances past the newline delimiter
                yield obj
        except JSONDecodeError:
            # finally arrived at the end of the string (or actually got malformed json from s3, supposed not to)
            pass

    @property
    def serialization(self):
        return {'JSON': {'RecordDelimiter': '\n'}}

    def combine(self, results: Sequence[Iterator[JsonValue]]) -> Iterator[JsonValue]:
        for r in results:
            yield from r



import unittest

from s3access.sql import AND, OR, EQ, NEQ, GTE, LT


class S3PathTestCase(unittest.TestCase):
    def test_sql_and(self):
        condition = AND(OR(EQ(7), NEQ(None)), GTE(2), LT(10))
        self.assertEqual("((foo = 7) OR (foo IS NOT NULL)) AND (foo >= 2) AND (foo < 10)",
                         condition.get_sql_fragment("foo"))

    def test_check_and(self):
        condition = AND(OR(EQ(7), NEQ(None)), GTE(2), LT(10))
        self.assertFalse(condition.check(None))
        self.assertFalse(condition.check(10))
        self.assertTrue(condition.check(2))
        self.assertTrue(condition.check(7))


if __name__ == '__main__':
    unittest.main()



import unittest

from s3access.s3path import S3Path


class S3PathTestCase(unittest.TestCase):
    def test_s3path(self):
        self.assertEqual(
            S3Path('s3://bucket/key'),
            S3Path('s3a://bucket/key'),
        )
        self.assertEqual(
            S3Path('s3://bucket/key'),
            S3Path('s3://bucket/key/'),
        )

    def test_params(self):
        p = S3Path('s3://bucket/foo=bar')
        self.assertEqual({
            'foo': 'bar',
        }, p.params)

    def test_with_params(self):
        p = S3Path('s3://bucket/foo=bar')
        p2 = p.with_params(foo='qux')
        self.assertEqual({
            'foo': 'qux',
        }, p2.params)
        p3 = p.with_params(qux='quuz', klm='pqq')
        self.assertEqual({
            'foo': 'bar',
            'qux': 'quuz',
            'klm': 'pqq',
        }, p3.params)
        self.assertEqual(
            S3Path('bucket/foo=bar/qux=quuz/klm=pqq'),
            p3,
        )

    def test_append(self):
        p = S3Path('s3://bucket/foo=bar')
        p /= 'hey=yea'
        self.assertEqual(
            p,
            S3Path('s3n://bucket/foo=bar/hey=yea')
        )

    def test_get_url(self):
        p = S3Path('s3://bucket/key/')
        self.assertEqual('s3://bucket/key', p.get_url())
        self.assertEqual('s3a://bucket/key', p.get_url('s3a'))

    def test_s3path_str(self):
        p = S3Path('s3://bucket/')
        p2 = p.with_params(client='cl')
        self.assertEqual('s3://bucket/client=cl', str(p2))

    def test_readme_example(self):
        p = S3Path("s3://bucket/key/part=value")

        p1 = p.with_params(foo=7, bar=29)
        self.assertEqual('s3://bucket/key/part=value/foo=7/bar=29', str(p1))

        p2 = p.with_params(foo=7, part='else', bar=29)
        self.assertEqual('s3://bucket/key/part=else/foo=7/bar=29', str(p2))

        p3 = p.with_bucket('other')
        self.assertEqual('s3://other/key/part=value', str(p3))

        p4 = p.with_key('path')
        self.assertEqual('s3://bucket/path', str(p4))


if __name__ == '__main__':
    unittest.main()



import unittest

from s3access.sql import quote


class S3PathTestCase(unittest.TestCase):
    def test_quote(self):
        self.assertEqual("3", quote(3))
        self.assertEqual("'3'", quote('3'))
        self.assertEqual("'foo'", quote('foo'))
        self.assertEqual("'foo'''", quote("foo'"))
        self.assertEqual("(1, 2, 3)", quote([1, 2, 3]))
        self.assertEqual("(1, 2, 3)", quote((1, 2, 3)))
        self.assertEqual("(1)", quote({1}))
        self.assertEqual("('foo', '''bar''')", quote(("foo", "'bar'")))
        self.assertEqual("NULL", quote(None))


if __name__ == '__main__':
    unittest.main()



import re

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

    start_result = re.search(r"<!-- START doctoc [^>]+-->", long_description)
    if start_result:
        pos = start_result.span(0)[0]
        before = long_description[:pos]
        remaining = long_description[pos:]
        end_result = re.search(r"<!-- END doctoc [^>]+-->", remaining)
        end_pos = end_result.span(0)[1]
        after = remaining[end_pos:]
        long_description = before + after

__pkginfo__ = {}
with open("s3access/__pkginfo__.py") as fh:
    exec(fh.read(), __pkginfo__)

install_requires = [
    'readstr>=0.5.0',
    'boto3>=1.17.0',  # first version supporting select_object_content (s3 select)
]

async_select_requires = [
    'aiobotocore>=1.3.0',
]

pandas_requires = [
    'pandas>=1.0.0',
    'pyarrow>=3.0.0',
]

tests_require = ['pytest'] + install_requires + async_select_requires + pandas_requires

setuptools.setup(
    name="s3access",
    version=__pkginfo__['__version__'],
    author="Julian Fleischer",
    author_email="tirednesscankill@warhog.net",
    description="Access Parquet files in S3 via S3 Select",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/scravy/s3access",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    extras_require={
        'async': async_select_requires,
        'pandas': pandas_requires,
        'dev': tests_require,
    },
    python_requires='>=3.8',
    tests_require=tests_require,
    install_requires=install_requires,
)



import unittest
import pandas as pd
import numpy as np
import tempfile
from datetime import date, datetime

from s3access.s3pandas.reader import Pandas


class StrictPandasReader(unittest.TestCase):

    def test_categorical_combine(self):
        a = b"a\nb\n"
        b = b"b\nc\n"
        columns = {'a': 'category'}
        reader = Pandas(strict=True)
        a_df = reader.read(a, columns)
        self.assertIn('a', a_df.columns)
        self.assertIsInstance(a_df['a'].dtype, pd.CategoricalDtype)
        b_df = reader.read(b, columns)
        self.assertIn('a', b_df.columns)
        self.assertIsInstance(b_df['a'].dtype, pd.CategoricalDtype)

        c_df = reader.combine([a_df, b_df])
        self.assertIn('a', c_df.columns)
        self.assertIsInstance(c_df['a'].dtype, pd.CategoricalDtype)

    def test_date_parsing(self):
        a = b"2021-06-05,2021-06-05T00:10:10\n2021-06-01,2021-06-01T23:59:59"

        columns = {'a': date, 'b': datetime}
        reader = Pandas(strict=True)
        a_df = reader.read(a, columns)
        self.assertIn('a', a_df.columns)
        self.assertEqual(a_df['a'].dtype, np.dtype('<M8[ns]'))
        self.assertEqual(a_df['b'].dtype, np.dtype('<M8[ns]'))

    def test_categorical_write_support(self):
        # some version of pyarrow did not support saving categorical data
        a = pd.DataFrame({"a": ['a', 'b', 'a', 'a', 'c', 'd', 'a', 'b', 'c', 'd']})
        a['a'] = a['a'].astype('category')

        with tempfile.TemporaryFile() as fp:
            a.to_parquet(fp)
            fp.seek(0)
            restored = pd.read_parquet(fp)
        self.assertIn('a', restored.columns)
        self.assertIsInstance(restored['a'].dtype, pd.CategoricalDtype)
        self.assertTrue(np.all(a['a'] == restored['a']))
